{********************************************************}
{                                                        }
{                 Zeos Database Objects                  }
{            Interbase Query and Table components        }
{                                                        }
{       Copyright (c) 1999-2001 Sergey Seroukhov         }
{                                                        }
{********************************************************}

unit ZIbSqlQuery;

interface

{$R *.DCR}

uses
  SysUtils, Windows, Db, Classes, ZDirSql, ZDirIbSql, DbCommon,
  ZIbSqlCon, ZIbSqlTr, ZToken, ZLibIbSql, ZSqlExtra, ZQuery,
  ZSqlTypes, ZSqlItems;

{$INCLUDE ..\Zeos.inc}

type

  TZIbSqlOption = (ioAutoIncKey);
  TZIbSqlOptions = set of TZIbSqlOption;

  { Direct Interbase dataset with descendant of TZDataSet }
  TZCustomIbSqlDataset = class(TZDataSet)
  private
    FExtraOptions: TZIbSqlOptions;
    procedure SetDatabase(Value: TZIbSqlDatabase);
    procedure SetTransact(Value: TZIbSqlTransact);
    function  GetDatabase: TZIbSqlDatabase;
    function  GetTransact: TZIbSqlTransact;
  protected
    { Overriding ZDataset methods }
    procedure QueryRecord; override;
    procedure UpdateAfterInit(RecordData: PRecordData); override;
    {$IFDEF WITH_IPROVIDER}
    { IProvider support }
    function  PSInTransaction: Boolean; override;
    function  PSExecuteStatement(const ASql: string; AParams: TParams;
              ResultSet: Pointer): Integer; override;
    procedure PSSetCommandText(const CommandText: string); override;
    {$ENDIF}
  public
    constructor Create(AOwner: TComponent); override;

    procedure AddTableFields(Table: string; SqlFields: TSqlFields); override;
    procedure AddTableIndices(Table: string; SqlFields: TSqlFields;
      SqlIndices: TSqlIndices); override;
  published
    property ExtraOptions: TZIbSqlOptions read FExtraOptions write FExtraOptions;
    property Database: TZIbSqlDatabase read GetDatabase write SetDatabase;
    property Transaction: TZIbSqlTransact read GetTransact write SetTransact;
  end;

{ Direct IbSql query with descendant of TDataSet }
  TZIbSqlQuery = class(TZCustomIbSqlDataset)
  public
    property ParamCount;
  published
    property Params;
    property ParamCheck;
    property DataSource;

    property Sql;
    property RequestLive;
    property Database;
    property Transaction;
    property Active;
  end;

{ Direct IbSql query with descendant of TDataSet }
  TZIbSqlTable = class(TZCustomIbSqlDataset)
  published
    property TableName;
    property ReadOnly;
    property Database;
    property Transaction;
    property Active;
  end;

implementation

uses ZExtra, ZDBaseConst, ZBlobStream, Math;

{********** TZCustomIbSqlDataset implementation **********}

{ Class constructor }
constructor TZCustomIbSqlDataset.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Query := TDirIbSqlQuery.Create(nil, nil);
  DatabaseType := dtInterbase;
  FExtraOptions := [];
end;

{ Set connect to database component }
procedure TZCustomIbSqlDataset.SetDatabase(Value: TZIbSqlDatabase);
begin
  inherited SetDatabase(Value);
end;

{ Set connect to transact-server component }
procedure TZCustomIbSqlDataset.SetTransact(Value: TZIbSqlTransact);
begin
  inherited SetTransact(Value);
end;

{ Get connect to database component }
function TZCustomIbSqlDataset.GetDatabase: TZIbSqlDatabase;
begin
  Result := TZIbSqlDatabase(DatabaseObj);
end;

{ Get connect to transact-server component }
function TZCustomIbSqlDataset.GetTransact: TZIbSqlTransact;
begin
  Result := TZIbSqlTransact(TransactObj);
end;

{ Read query from server to internal buffer }
procedure TZCustomIbSqlDataset.QueryRecord;
var
  I, Count: Integer;
  RecordData: PRecordData;
  FieldDesc: PFieldDesc;
  TempStr: string;
  TempLong: LongInt;
  TempDouble: Double;
  TempTime: TDateTime;
  TimeStamp: TTimeStamp;
  TempDate: TCTimeStructure;
  TempPtr: PISC_QUAD;
  BlobPtr: PRecordBlob;
  Cancel: Boolean;
begin
  Count := SqlBuffer.Count;
  while not Query.EOF and (Count = SqlBuffer.Count) do
  begin
    { Go to the record }
    if SqlBuffer.FillCount > 0 then
      Query.Next;
    { Invoke OnProgress event }
    if Assigned(OnProgress) then
    begin
      Cancel := False;
      OnProgress(Self, psRunning, ppFetching, Query.RecNo+1,
        MaxIntValue([Query.RecNo+1, Query.RecordCount]), Cancel);
      if Cancel then Query.Close;
    end;
    if Query.EOF then Break;
    { Getting record }
    RecordData := SqlBuffer.Add;
    for I := 0 to SqlBuffer.SqlFields.Count - 1 do
    begin
      FieldDesc := SqlBuffer.SqlFields[I];
      if FieldDesc.FieldNo < 0 then Continue;
      if Query.FieldIsNull(FieldDesc.FieldNo) and
        not (FieldDesc.FieldType in [ftBlob, ftMemo]) then
        Continue;

      case FieldDesc.FieldType of
        ftString:
          begin
            case Query.FieldType(FieldDesc.FieldNo) of
              SQL_VARYING:
                begin
                  SqlBuffer.SetFieldDataLen(FieldDesc,
                    Pointer(LongInt(Query.FieldBuffer(FieldDesc.FieldNo)) + 2),
                    RecordData, Query.FieldSize(FieldDesc.FieldNo));
                  RecordData.Bytes[FieldDesc.Offset
                    + PSmallInt(Query.FieldBuffer(FieldDesc.FieldNo))^ + 1] := 0;
                end;
              else
                begin
                  TempStr := Query.Field(FieldDesc.FieldNo);
                  SqlBuffer.SetFieldDataLen(FieldDesc,
                    PChar(TempStr), RecordData, Length(TempStr));
                end;
            end;
          end;
        ftInteger, ftSmallInt:
          begin
            SqlBuffer.SetFieldData(FieldDesc, Query.FieldBuffer(FieldDesc.FieldNo),
              RecordData);
          end;
        ftFloat:
          begin
            case Query.FieldType(FieldDesc.FieldNo) of
              SQL_LONG:
                TempDouble := PLongInt(Query.FieldBuffer(FieldDesc.FieldNo))^
                  / IntPower(10, Query.FieldDecimals(FieldDesc.FieldNo));
              SQL_DOUBLE:
                TempDouble := PDouble(Query.FieldBuffer(FieldDesc.FieldNo))^;
              else
                TempDouble := PSingle(Query.FieldBuffer(FieldDesc.FieldNo))^;
            end;
            SqlBuffer.SetFieldData(FieldDesc, @TempDouble, RecordData);
          end;
        ftDateTime:
          begin
            isc_decode_date(PISC_QUAD(Query.FieldBuffer(FieldDesc.FieldNo)), @TempDate);
            TimeStamp := DateTimeToTimeStamp(EncodeDate(TempDate.tm_year + 1900,
              TempDate.tm_mon+1, TempDate.tm_mday) + EncodeTime(TempDate.tm_hour,
              TempDate.tm_min, TempDate.tm_sec, 0));
            TempTime := TimeStampToMSecs(TimeStamp);
            SqlBuffer.SetFieldData(FieldDesc, @TempTime, RecordData);
          end;
        ftDate:
          begin
            isc_decode_date(PISC_QUAD(Query.FieldBuffer(FieldDesc.FieldNo)), @TempDate);
            TempDate.tm_hour := 0;
            TempDate.tm_min := 0;
            TempDate.tm_sec := 0;
            TempLong := DateTimeToTimeStamp(EncodeDate(TempDate.tm_year + 1900,
              TempDate.tm_mon+1, TempDate.tm_mday) + EncodeTime(TempDate.tm_hour,
              TempDate.tm_min, TempDate.tm_sec, 0)).Date;
            SqlBuffer.SetFieldData(FieldDesc, @TempLong, RecordData);
          end;
        ftBlob, ftMemo:
          begin
            { Initialize blob field }
            BlobPtr := PRecordBlob(@RecordData.Bytes[FieldDesc.Offset+1]);
            BlobPtr.BlobType := btExternal;
            BlobPtr.Data := nil;
            BlobPtr.Size := 0;
            TempPtr := PISC_QUAD(Query.FieldBuffer(FieldDesc.FieldNo));
            { Fill not null fields }
            if Assigned(TempPtr) then
            begin
              RecordData.Bytes[FieldDesc.Offset] := 0;
              BlobPtr.Handle.Ptr := TempPtr.gds_quad_high;
              BlobPtr.Handle.PtrEx := TempPtr.gds_quad_Low;
            end
            { Fill null fields }
            else
            begin
              BlobPtr.Handle.Ptr := 0;
              BlobPtr.Handle.PtrEx := 0;
            end;
          end;
        else
          DatabaseError(SUnknownType + FieldDesc.Alias);
      end;
    end;
    { Filter received record }
    SqlBuffer.FilterItem(SqlBuffer.Count-1);
  end;
end;

{************** Sql-queries processing ******************}

{ Fill collection with fields }
procedure TZCustomIbSqlDataset.AddTableFields(Table: string;
  SqlFields: TSqlFields);
var
  Size: Integer;
  Decimals: Integer;
  FieldType: TFieldType;
  Query: TDirIbSqlQuery;
  Default: string;
  BlobType: TBlobType;
begin
  Query := TDirIbSqlQuery(Transaction.QueryHandle);
  Query.ShowColumns(Table, '');
  while not Query.EOF do
  begin
    { Evalute field parameters }
    Size := StrToIntDef(Query.Field(3), 0);
    Decimals := StrToIntDef(Query.Field(6), 0);
    FieldType := IbSqlToDelphiType(StrToIntDef(Query.Field(2), 0), Decimals);
    if FieldType = ftBlob then BlobType := btExternal
    else BlobType := btInternal;
    if FieldType <> ftString then Size := 0;
    Default := Query.Field(5);
    StrTok(Default, ' '#9#10#13);

    { Put new field description }
    SqlFields.Add(Table, Query.Field(1), '', '', FieldType, Size, Decimals,
      atNone, Query.Field(4) <> '1', Default, BlobType);
    Query.Next;
  end;
  Query.Close;
end;

{ Fill collection with indices }
procedure TZCustomIbSqlDataset.AddTableIndices(Table: string;
  SqlFields: TSqlFields; SqlIndices: TSqlIndices);
var
  KeyType: TKeyType;
  SortType: TSortType;
  Query: TDirIbSqlQuery;
begin
  Query := TDirIbSqlQuery(TransactObj.QueryHandle);
  Query.ShowIndexes(Table);
  while not Query.EOF do
  begin
    { Define a key type }
    if Query.Field(3) = '1' then
    begin
      if StrCmpBegin(Query.Field(1),'RDB$PRIMARY') then
        KeyType := ktPrimary
      else KeyType := ktUnique;
    end else KeyType := ktIndex;
    { Define a sorting mode }
    if Query.Field(4) <> '1' then SortType := stAscending
    else SortType := stDescending;

    { Put new index description }
    SqlIndices.AddIndex(Query.Field(1), Table, Query.Field(5), KeyType, SortType);
    Query.Next;
  end;
  Query.Close;
end;

{ Update record after initialization }
procedure TZCustomIbSqlDataset.UpdateAfterInit(RecordData: PRecordData);

  function FindPrimaryKey: PFieldDesc;
  var
    I: Integer;
    IndexDesc: PIndexDesc;
  begin
    Result := nil;
    if SqlParser.Tables.Count = 0 then Exit;
    { Find primary key }
    IndexDesc := nil;
    for I := 0 to SqlBuffer.SqlIndices.Count-1 do
      if StrCaseCmp(SqlBuffer.SqlIndices[I].Table, SqlParser.Tables[0])
        and (SqlBuffer.SqlIndices[I].KeyType = ktPrimary) then
      begin
        IndexDesc := SqlBuffer.SqlIndices[I];
        Break;
      end;
    { Check primary key }
    if (IndexDesc = nil) or (IndexDesc.FieldCount <> 1) then Exit;
    Result := SqlBuffer.SqlFields.FindByName(SqlParser.Tables[0],
      IndexDesc.Fields[0]);
    if Result = nil then Exit;
    if Result.FieldType <> ftInteger then
      Result := nil;
  end;

var
  FieldDesc: PFieldDesc;
begin
  inherited UpdateAfterInit(RecordData);

  if ioAutoIncKey in FExtraOptions then
  begin
    FieldDesc := FindPrimaryKey;
    if FieldDesc <> nil then
      SqlBuffer.SetField(FieldDesc, EvaluteDef(Format('GEN_ID(%s_%s_gen, 1)',
        [SqlParser.Tables[0], FieldDesc.Field])), RecordData);
  end;
end;

{$IFDEF WITH_IPROVIDER}
{ IProvider support }

{ Is in transaction }
function TZCustomIbSqlDataset.PSInTransaction: Boolean;
begin
  Result := True;
end;

{ Execute an sql statement }
function TZCustomIbSqlDataset.PSExecuteStatement(const ASql: string; AParams: TParams;
  ResultSet: Pointer): Integer;
begin
  if Assigned(ResultSet) then
  begin
    TDataSet(ResultSet^) := TZIbSqlQuery.Create(nil);
    with TZIbSqlQuery(ResultSet^) do
    begin
      Sql.Text := ASql;
      Params.Assign(AParams);
      Open;
      Result := RowsAffected;
    end;
  end else
    Result := TransactObj.ExecSql(ASql);
end;

{ Set command query }
procedure TZCustomIbSqlDataset.PSSetCommandText(const CommandText: string);
begin
  Close;
  if Self is TZIbSqlQuery then
    TZIbSqlQuery(Self).Sql.Text := CommandText
  else
  if Self is TZIbSqlTable then
    TZIbSqlQuery(Self).TableName := CommandText;
end;

{$ENDIF}

end.
