{********************************************************}
{                                                        }
{                 Zeos Database Objects                  }
{              MySql Transaction component               }
{                                                        }
{       Copyright (c) 1999-2001 Sergey Seroukhov         }
{                                                        }
{********************************************************}

unit ZPgSqlTr;

interface

{$R *.DCR}

uses
  Classes, ZDirPgSql, ZPgSqlCon, ZTransact, ZLibPgSql, ZSqlTypes;

{$INCLUDE ..\Zeos.inc}

type
  { PostgreSql transaction }
  TZPgSqlTransact = class(TZTransact)
  private
    function  GetDatabase: TZPgSqlDatabase;
    function GetPID: Integer;
    procedure SetDatabase(Value: TZPgSqlDatabase);
    function GetLastInsertOid: Oid;
  public
    constructor Create(AOwner: TComponent); override;

    procedure Connect; override;
    procedure Recovery(Force: Boolean); override;

    procedure AddMonitor(Monitor: TZMonitor); override;
    procedure DeleteMonitor(Monitor: TZMonitor); override;

    procedure Reset;

    property PID: Integer read GetPID;
    property LastInsertOid: Oid read GetLastInsertOid;
  published
    property Database: TZPgSqlDatabase read GetDatabase write SetDatabase;
    property AutoRecovery;
    property TransactSafe;
    property TransIsolation;
  end;

  { PostgreSQL class for asynchronous notifying}
  TZPgSqlNotify = class(TZNotify)
  private
    function GetTransact: TZPgSqlTransact;
  protected
    procedure SetTransact(Value: TZPgSqlTransact);
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Transaction: TZPgSqlTransact read GetTransact write SetTransact;
  end;
  
implementation

{***************** TZPgSqlTransact implementation *****************}

{ Class constructor }
constructor TZPgSqlTransact.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FHandle := TDirPgSqlTransact.Create(nil);
  FQuery  := TDirPgSqlQuery.Create(nil, TDirPgSqlTransact(FHandle));
  AutoRecovery := True;
  FDatabaseType := dtPostgreSql;
end;

{ Get database component }
function TZPgSqlTransact.GetDatabase: TZPgSqlDatabase;
begin
  Result := TZPgSqlDatabase(FDatabase);
end;

{ Set database component }
procedure TZPgSqlTransact.SetDatabase(Value: TZPgSqlDatabase);
begin
  inherited SetDatabase(Value);
end;

{ Retrieve backend server's process id (PID }
function TZPgSqlTransact.GetPID: Integer;
begin
  Result := TDirPgSqlTransact(Handle).PID;
end;

{ Get last inserted oid value }
function TZPgSqlTransact.GetLastInsertOid: Oid;
begin
  Result := TDirPgSqlQuery(FQuery).LastInsertOid;
end;

{ Recovery after errors }
procedure TZPgSqlTransact.Recovery(Force: Boolean);
begin
  if AutoRecovery or Force then
    FHandle.Rollback;
end;

{ Connect to Sql-database }
procedure TZPgSqlTransact.Connect;
begin
  inherited Connect;
  ExecSql('SET DateStyle TO ''ISO''');
end;

{ Add monitor into monitor list }
procedure TZPgSqlTransact.AddMonitor(Monitor: TZMonitor);
begin
  ZDirPgSql.MonitorList.AddMonitor(Monitor);
end;

{ Delete monitor from monitor list }
procedure TZPgSqlTransact.DeleteMonitor(Monitor: TZMonitor);
begin
  ZDirPgSql.MonitorList.DeleteMonitor(Monitor);
end;

{ Reconnect to the database server}
procedure TZPgSqlTransact.Reset;
begin
  TDirPgSqlTransact(Handle).Reset;
end;

{***************** TZPgSqlNotify implementation *****************}

{ Class constructor }
constructor TZPgSqlNotify.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FHandle := TDirPgSqlNotify.Create(nil, nil);
end;

{ Get transact object }
function TZPgSqlNotify.GetTransact: TZPgSqlTransact;
begin
  Result := TZPgSqlTransact(FTransact);
end;

{ Set transact object }
procedure TZPgSqlNotify.SetTransact(Value: TZPgSqlTransact);
begin
  inherited SetTransact(Value);
end;

end.
